#! /usr/bin/env/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 1999-2015 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2015 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


import types

import pytest

from ll.xist import xsc, xnd, sims
from ll.xist.scripts import dtd2xsc


def dtd2mod(s, shareattrs=None, **kwargs):
	xnd = dtd2xsc.urls2xnd([s], encoding="utf-8", model="fullonce", **kwargs)

	if shareattrs is not None:
		xnd.shareattrs(shareattrs)

	mod = types.ModuleType("test")
	mod.__file__ = "test.py"

	code = str(xnd)
	print("Module source generated from DTDs:")
	print(code)
	code = compile(code, "test.py", "exec")
	exec(code, mod.__dict__)
	return mod


@pytest.mark.lxml
def test_convert():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo (bar+)>
	<!ATTLIST foo
		id    ID    #IMPLIED
		xmlns CDATA #FIXED "http://xmlns.foo.com/foo"
	>
	<!ELEMENT bar EMPTY>
	<!ATTLIST bar
		bar1 CDATA               #REQUIRED
		bar2 (bar2)              #IMPLIED
		bar3 (bar3a|bar3b|bar3c) #IMPLIED
		bar-4 (bar-4a|bar-4b)    #IMPLIED
		bar_4 (bar_4a|bar_4b)    #IMPLIED
		bar_42 (bar_42a|bar_42b) #IMPLIED
		class CDATA              #IMPLIED
		foo:bar CDATA            #IMPLIED
	>
	"""
	ns = dtd2mod(dtdstring)

	assert ns.foo.xmlns == "http://xmlns.foo.com/foo"
	assert isinstance(ns.foo.model, sims.Elements)
	assert len(ns.foo.model.elements) == 1
	assert ns.foo.model.elements[0] == ns.bar
	assert issubclass(ns.foo.Attrs.id, xsc.IDAttr)
	assert "xmlns" not in ns.foo.Attrs
	assert isinstance(ns.bar.model, sims.Empty)

	assert "bar" not in ns.bar.Attrs

	assert issubclass(ns.bar.Attrs.bar1, xsc.TextAttr)
	assert ns.bar.Attrs.bar1.required is True

	assert issubclass(ns.bar.Attrs.bar2, xsc.BoolAttr)
	assert ns.bar.Attrs.bar2.required is False

	assert issubclass(ns.bar.Attrs.bar3, xsc.TextAttr)
	assert ns.bar.Attrs.bar3.required is False
	assert ns.bar.Attrs.bar3.values == ("bar3a", "bar3b", "bar3c")

	# Attributes are alphabetically sorted
	assert issubclass(ns.bar.Attrs.bar_4, xsc.TextAttr)
	assert ns.bar.Attrs.bar_4.xmlname == "bar-4"
	assert ns.bar.Attrs.bar_4.values == ("bar-4a", "bar-4b")

	assert issubclass(ns.bar.Attrs.bar_42, xsc.TextAttr)
	assert ns.bar.Attrs.bar_42.xmlname == "bar_4"
	assert ns.bar.Attrs.bar_42.values == ("bar_4a", "bar_4b")

	assert issubclass(ns.bar.Attrs.bar_422, xsc.TextAttr)
	assert ns.bar.Attrs.bar_422.xmlname == "bar_42"
	assert ns.bar.Attrs.bar_422.values == ("bar_42a", "bar_42b")


@pytest.mark.lxml
def test_charref():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo EMPTY>
	<!ENTITY bar "&#xff;">
	"""
	ns = dtd2mod(dtdstring)

	assert ns.bar.codepoint == 0xff


@pytest.mark.lxml
def test_keyword():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo EMPTY>
	<!ATTLIST foo
		class CDATA              #IMPLIED
	>
	"""
	ns = dtd2mod(dtdstring)
	assert issubclass(ns.foo.Attrs.class_, xsc.TextAttr)
	assert ns.foo.Attrs.class_.__name__ == "class_"
	assert ns.foo.Attrs.class_.xmlname == "class"


@pytest.mark.lxml
def test_quotes():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo EMPTY>
	"""
	ns = dtd2mod(dtdstring, defaultxmlns='"')
	assert ns.foo.xmlns == '"'


@pytest.mark.lxml
def test_unicode():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo EMPTY>
	"""
	ns = dtd2mod(dtdstring, defaultxmlns='\u3042')
	assert ns.foo.xmlns == '\u3042'


@pytest.mark.lxml
def test_unicodequotes():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo EMPTY>
	"""
	ns = dtd2mod(dtdstring, defaultxmlns='"\u3042"')
	assert ns.foo.xmlns == '"\u3042"'


@pytest.mark.lxml
def test_badelementname():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT class EMPTY>
	"""
	ns = dtd2mod(dtdstring)
	assert issubclass(ns.class_, xsc.Element)


@pytest.mark.lxml
def test_shareattrsnone():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo (bar)>
	<!ATTLIST foo
		baz CDATA              #IMPLIED
	>
	<!ELEMENT bar EMPTY>
	<!ATTLIST bar
		baz CDATA              #IMPLIED
	>
	"""
	ns = dtd2mod(dtdstring, shareattrs=None)
	assert not hasattr(ns, "baz")


@pytest.mark.lxml
def test_shareattrsdupes():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo (bar)>
	<!ATTLIST foo
		baz  CDATA             #IMPLIED
		baz2 CDATA             #IMPLIED
	>
	<!ELEMENT bar EMPTY>
	<!ATTLIST bar
		baz  CDATA             #IMPLIED
		baz2 CDATA             #REQUIRED
	>
	"""
	ns = dtd2mod(dtdstring, shareattrs=False)
	assert issubclass(ns.foo.Attrs.baz, ns.baz.baz)
	assert issubclass(ns.bar.Attrs.baz, ns.baz.baz)
	assert not hasattr(ns, "baz2")
	assert not ns.foo.Attrs.baz2.required
	assert ns.bar.Attrs.baz2.required


@pytest.mark.lxml
def test_shareattrsall():
	dtdstring = """<?xml version='1.0' encoding='us-ascii'?>
	<!ELEMENT foo (bar)>
	<!ATTLIST foo
		baz  CDATA             #IMPLIED
		bazz CDATA             #IMPLIED
	>
	<!ELEMENT bar EMPTY>
	<!ATTLIST bar
		baz  CDATA             #IMPLIED
		bazz CDATA             #REQUIRED
	>
	"""
	ns = dtd2mod(dtdstring, shareattrs=True)
	assert issubclass(ns.foo.Attrs.baz, ns.baz.baz)
	assert issubclass(ns.bar.Attrs.baz, ns.baz.baz)

	assert not ns.foo.Attrs.bazz.required
	assert ns.bar.Attrs.bazz.required



#!/usr/local/bin/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 2007-2015 by LivingLogic AG, Bayreuth/Germany
## Copyright 2007-2015 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


"""
Purpose
-------

``ucat`` is a script for printing files. It is an URL-enabled version of the
``cat`` system command. Via :mod:`ll.url` and :mod:`ll.orasql` ``ucat`` supports
``ssh`` and ``oracle`` URLs.


Options
-------

``ucat`` supports the following options:

	``urls``
		One or more URLs to be printed.

	``-r``, ``--recursive`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Prints directory content recursively.

	``-x``, ``--ignoreerrors`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Ignores file i/o errors occuring during the output process (otherwise
		the script will be aborted).

	``-i``, ``--include`` : pattern(s)
		Only print files whose name matches one of the specified patterns.

	``-e``, ``--exclude`` : pattern(s)
		Don't print files whose name matches one of the specified patterns.

	``--enterdir`` : pattern(s)
		Only enter directories whose name matches one of the specified patterns.

	``--skipdir`` : pattern(s)
		Don't enter directories whose name matches one of the specified patterns.


Examples
--------
Print a file::

	$ ucat foo.txt

Print a remote file::

	$ ucat ssh://user@www.example.org/~/foo.txt

Print the SQL source code of the procedure ``FOO`` in an Oracle database::

	$ ucat oracle://user:pwd@oracle.example.org/procedure/FOO

"""


import sys, re, argparse, contextlib, errno

from ll import misc, url

try:
	import astyle
except ImportError:
	from ll import astyle

try:
	from ll import orasql # activate the oracle scheme
except ImportError:
	pass


__docformat__ = "reStructuredText"


def main(args=None):
	def catone(urlread):
		if urlread.isdir():
			if args.recursive:
				for u in urlread.walkfiles(include=args.include, exclude=args.exclude, enterdirs=args.enterdir, skipdirs=args.skipdir):
					catone(urlread/u)
			else:
				raise IOError(errno.EISDIR, "Is a directory", str(urlread))
		else:
			try:
				with contextlib.closing(urlread.open("rb")) as fileread:
					size = 0
					while True:
						data = fileread.read(262144)
						if data:
							sys.stdout.buffer.write(data)
						else:
							break
			except Exception:
				if not args.ignoreerrors:
					raise

	p = argparse.ArgumentParser(description="print URL content on the screen", epilog="For more info see http://www.livinglogic.de/Python/scripts/ucat.html")
	p.add_argument("urls", metavar="url", help="URLs to be printed", nargs="+", type=url.URL)
	p.add_argument("-r", "--recursive", dest="recursive", help="Print stuff recursively? (default: %(default)s)", action=misc.FlagAction, default=False)
	p.add_argument("-x", "--ignoreerrors", dest="ignoreerrors", help="Ignore errors? (default: %(default)s)", action=misc.FlagAction, default=False)
	p.add_argument("-i", "--include", dest="include", metavar="PATTERN", help="Include only URLs matching PATTERN", action="append")
	p.add_argument("-e", "--exclude", dest="exclude", metavar="PATTERN", help="Exclude URLs matching PATTERN", action="append")
	p.add_argument(      "--enterdir", dest="enterdir", metavar="PATTERN", help="Only enter directories matching PATTERN", action="append")
	p.add_argument(      "--skipdir", dest="skipdir", metavar="PATTERN", help="Skip directories matching PATTERN", action="append")

	args = p.parse_args(args)
	with url.Context():
		for u in args.urls:
			catone(u)


if __name__ == "__main__":
	sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

"""
Purpose
=======

``rul4`` is a script that can be used for rendering UL4 templates. Templates
have access to Oracle, MySQL, SQLite and Redis databases and can execute system
commands.


Options
=======

``rul4`` supports the following options:

	``templates``
		One or more template files. A file named ``-`` will be treated as
		standard input. The first file in the list is the main template, i.e. the
		one that gets rendered. All templates will be available in the main
		template as the ``templates`` dictionary. The keys are the base names
		of the files (i.e. ``foo.ul4`` will be ``templates.foo``; stdin will be
		``templates.stdin``).

	``--oracle`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Provide the object ``oracle`` to the template or not (see below)?

	``--sqlite`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Provide the object ``sqlite`` to the template or not (see below)?

	``--mysql`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Provide the object ``mysql`` to the template or not (see below)?

	``--redis`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Provide the object ``redis`` to the template or not (see below)?

	``--system`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Provide the object ``system`` to the template or not (see below)?

	``--import`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Provide the function ``import`` to the template or not (see below)?

	``-e``, ``--encoding``
		The encoding of the templates files (default ``utf-8``)

	``-w``, ``--whitespace`` : ``keep``, ``strip``, or ``smart``
		Specifies how to handle whitespace in the template

	``-D``, ``--define``
		Defines an additional value that will be available inside the template.
		``-D`` can be specified multiple times. The following formats are supported:

			``var``
				Defines ``var`` as an empty string;

			``var=value``
				Defines ``var`` as the string ``value``;

			``var:type``
				Defines ``var`` as an empty variable of the type ``type``;

			``var:type=value``
				Defines ``var`` as a variable of the type ``type`` with the value
				``value``.

		``type`` can be any of the following:

			``int``
				``value`` is an integer value.

			``float``
				``value`` is a float value.

			``bool``
				``value`` is a boolean value. ``0``, ``no``, ``false``, ``False`` or
				the empty string will be recognized as false and ``1``, ``yes``,
				``true`` or ``True`` will be recognized as true.

			``str``
				``value`` is a string.

			``oracle``
				``value`` will be a connection to an Oracle database, e.g.::

					-Ddb:oracle=user/pwd@database

			``sqlite``
				``value`` is a connection to an SQLite database.

			``mysql``
				``value`` is a connection to a MySQL database.

			``redis``
				``value`` will be a connection to an Redis database, e.g.::

					-Ddb:redis=192.168.123.1:6379/42

				The port (i.e. the ``6379`` in the above value) is optional and
				defaults to 6379. The database number (i.e. the ``42`` in the above
				value) is also optional and defaults to 0.


Template variables
==================

Inside the template the following variables are available (if enabled via
the matching options):

	``templates``
		A dictionary containing all the templates specified on the command line.

	``encoding``
		The output encoding.

	``system``
		An object with an ``execute`` method that executes system commands and
		returns their output, e.g. the template::

			<?print system.execute("whoami")?>

		will output the user name.

	``oracle``
		An object with a ``connect`` method that returns a connection to an oracle
		database.

	``mysql``
		An object with a ``connect`` method that return a MySQL connection for
		the MySQL connect strings passed in. A MySQL connect string is a string
		of the form ``user/pwd@host/db``.

	``sqlite``
		An object with a ``connect`` method that return a SQLite connection for
		the connect strings passed in. The connect string will be passed directly
		to :func:`sqlite3.connect`.

	``redis``
		An object with a ``connect`` method that return a Redis connection for
		the connect strings passed in. The connectstring is of the form
		``hostname:port/db``. ``port`` and ``db`` are optional.

	``import``
		``import`` is a function that read a file (containing an UL4 template)
		from disk and returns the compiled UL4 template. Its first parameter is
		the filename and its second parameter is the encoding of the file. The
		encoding parameter is optional and defaults to ``"utf-8"``::

			<?code template = import("/home/user/template/foo.ul4")?>

All variables defined via the :option:`-D`/:option:`--define` option will also
be available. (Note that you can't overwrite any of the predefined variables).


Database connections
--------------------

All connection objects (except ``redis``) have a ``query`` method that executes
the query passed in and returns an iterator over the resulting records. This
``query`` method requires at least one positional argument. Arguments alternate
between fragments of the SQL query and parameters that will be embedded in the
query. For example::

	<?code db = oracle.connect("user/pwd@db")?>
	<?code name = "Bob"?>
	<?for p in db.query("select * from person where firstname=", name, " or lastname=", name)?>
		...

The records returned from ``query`` are dict-like objects mapping field names to
field values.

Connection objects also have an ``execute`` method that supports the same
parameters as ``query`` but doesn't return an iterable result. This can be used
to call functions or procedures.

Calling functions or procedures with out parameters can be done with variable
objects that can be created with the methods :meth:`int`, :meth:`number`,
:meth:`str`, :meth:`clob` and :meth:`date`. The resulting value of the out
parameter is available from the ``value`` attribute of the variable object.
The following example creates a function, calls it to get at the result and
drops it again::

	<?code db = oracle.connect('user/pwd@database')?>
	<?code db.execute('''
		create or replace function ul4test(p_arg integer)
		return integer
		as
		begin
			return 2*p_arg;
		end;
	''')?>
	<?code vout = db.int()?>
	<?code db.execute('begin ', vout, ' := ul4test(42); end;')?>
	<?print vout.value?>
	<?code db.execute('drop function ul4test')?>


Redis connections have a ``get`` and a ``put`` method::

	<?code db = redis.connect("192.168.123.42/1")?>
	<?code value = db.get("key")?>
	<?if isnone(value)?>
		<?code value = "foobar"?>
		<?code db.put("key", value, timedelta(seconds=10*60))?>
	<?end if?>

The timeout value in the ``put`` method is optional. Without it the value will
be stored indefinitely.


Example
=======

This example shows how to connect to an Oracle database and output the content
of a ``person`` table into an XML file.

Suppose we have a database table that looks like this::

	create table person
	(
		id integer not null,
		firstname varchar2(200),
		lastname varchar2(200)
	);

Then we can use the following template to output the table into an XML file::

	<?xml version='1.0' encoding='utf-8'?>
	<?code db = oracle.connect("user/pwd@database')?>
	<persons>
		<?for p in db.query("select id, firstname, lastname from person order by 2, 1")?>
			<person id="<?printx p.id?>">
				<firstname><?printx p.firstname?></firstname>
				<lastname><?printx p.lastname?></lastname>
			</person>
		<?end for?>
	</persons>

If we put the template into the file ``person.ul4`` we can call ``rul4`` like
this::

	rul4 person.ul4 >person.xml

We could also pass the connection to our database via the ``-D`` option and
disallow the script to make any database connections itself or execute any
system commands::

	rul4 person.ul4 -Ddb:oracle=user/pwd@database --oracle=0 --sqlite=0 --mysql=0 --redis=0 --system=0 >person.xml

Then the template could use the Oracle connection object ``db`` directly.
"""


import sys, os, argparse, datetime, keyword

from ll import ul4c, misc


__docformat__ = "reStructuredText"


class System:
	ul4attrs = {"execute"}

	def execute(self, cmd):
		return os.popen(cmd).read()


def import_(filename, encoding="utf-8"):
	with open(filename, "r", encoding=encoding) as f:
		data = f.read()

	template = ul4c.Template(data)

	return template


class Var:
	ul4attrs = {"value"}

	def __init__(self, value=None):
		self.value = value

	def ul4setattr(self, name, value):
		# As ``ul4attrs`` only contains ``"value"``, we will never be called with any other name
		self.value = value

	@misc.notimplemented
	def makevar(c, cursor):
		pass


class Connection:
	ul4attrs = {"query", "execute", "int", "number", "str", "clob", "date"}

	def __init__(self, connection):
		self.connection = connection

	def _execute(self, cursor, queryparts):
		query = []
		params = {}
		vars = {}
		for (i, part) in enumerate(queryparts):
			if i % 2:
				name = "value{}".format((i+1)//2)
				if isinstance(part, Var):
					params[name] = part.makevar(cursor)
					vars[name] = part
				else:
					params[name] = part
				query.append(":" + name)
			else:
				query.append(part)
		cursor.execute("".join(query), **params)
		for (name, var) in vars.items():
			var.value = params[name].getvalue(0)

	def query(self, *queryparts):
		cursor = self.connection.cursor()
		self._execute(cursor, queryparts)
		return cursor

	def execute(self, *queryparts):
		cursor = self.connection.cursor()
		self._execute(cursor, queryparts)

	@misc.notimplemented
	def str(self, value=None):
		pass

	@misc.notimplemented
	def clob(self, value=None):
		pass

	@misc.notimplemented
	def int(self, value=None):
		pass

	@misc.notimplemented
	def number(self, value=None):
		pass

	@misc.notimplemented
	def date(self, value=None):
		pass


class OracleConnection(Connection):
	class IntVar(Var):
		def makevar(self, c):
			var = c.var(int)
			var.setvalue(0, self.value)
			return var

	class NumberVar(Var):
		def makevar(self, c):
			var = c.var(float)
			var.setvalue(0, self.value)
			return var

	class StrVar(Var):
		def makevar(self, c):
			var = c.var(str)
			var.setvalue(0, self.value)
			return var

	class CLOBVar(Var):
		def makevar(self, c):
			from ll import orasql
			var = c.var(orasql.CLOB)
			var.setvalue(0, self.value)
			return var

	class DateVar(Var):
		def makevar(self, c):
			var = c.var(datetime.datetime)
			var.setvalue(0, self.value)
			return var

	def __repr__(self):
		connectstring = "{}@{}".format(self.connection.username, self.connection.tnsentry)
		return "<{}.{} schema={!r} at {:#x}>".format(self.__class__.__module__, self.__class__.__qualname__, connectstring, id(self))

	def str(self, value=None):
		return self.StrVar(value)

	def clob(self, value=None):
		return self.CLOBVar(value)

	def int(self, value=None):
		return self.IntVar(value)

	def number(self, value=None):
		return self.NumberVar(value)

	def date(self, value=None):
		return self.DateVar(value)


class Oracle:
	ul4attrs = {"connect"}

	def connect(self, connectstring):
		from ll import orasql
		return OracleConnection(orasql.connect(connectstring, readlobs=True))


class SQLite:
	ul4attrs = {"connect"}

	def connect(self, connectstring):
		import sqlite3
		connection = sqlite3.connect(connectstring)
		class Row(sqlite3.Row):
			def __getitem__(self, key):
				if isinstance(key, str):
					key = key.encode("ascii")
				return sqlite3.Row.__getitem__(self, key)
		connection.row_factory = Row
		return Connection(connection)


class MySQL:
	ul4attrs = {"connect"}

	def connect(self, connectstring):
		import MySQLdb
		from MySQLdb import cursors
		(user, host) = connectstring.split("@")
		(user, passwd) = user.split("/")
		(host, db) = host.split("/")
		return Connection(MySQLdb.connect(user=user, passwd=passwd, host=host, db=db, use_unicode=True, cursorclass=cursors.DictCursor))


class RedisConnection:
	ul4attrs = {"get", "put"}

	def __init__(self, host, port, db):
		import redis
		self.connection = redis.StrictRedis(host=host, port=port, db=db, decode_responses=True)

	def get(self, key):
		return self.connection.get(key)

	def set(self, key, data, timeout=None):
		if timeout is None:
			self.connection.set(key, data)
		else:
			self.connection.setex(key, timeout, data)


class Redis:
	ul4attrs = {"connect"}

	def connect(self, connectstring):
		(hostport, _, db) = connectstring.partition("/")
		if not db:
			db = 0
		(host, _, port) = hostport.partition(":")
		if not port:
			port = 6379
		return RedisConnection(host=host, port=port, db=db)


# Instantiate all "handlers"
system = System()
oracle = Oracle()
sqlite = SQLite()
mysql = MySQL()
redis = Redis()


def fixname(name):
	newname = "".join(c for (i, c) in enumerate(name) if (c.isalnum() if i else c.isalpha()) or c == "_")
	while keyword.iskeyword(newname):
		newname += "_"
	return newname


def define(arg):
	(name, _, value) = arg.partition("=")
	(name, _, type) = name.partition(":")
	if any(c != "_" and not (c.isalnum() if i else c.isalpha()) for (i, c) in enumerate(name)):
		raise argparse.ArgumentTypeError("{!r} is not a legal variable name".format(name))

	if type == "int":
		if not value:
			return (name, 0)
		try:
			return (name,  int(value))
		except ValueError:
			raise argparse.ArgumentTypeError("{!r} is not a legal integer value".format(value))
	elif type == "float":
		if not value:
			return (name, 0.)
		try:
			return (name, float(value))
		except ValueError:
			raise argparse.ArgumentTypeError("{!r} is not a legal float value".format(value))
	elif type == "bool":
		if value in ("", "0", "no", "false", "False"):
			return (name, False)
		if value in ("1", "yes", "true", "True"):
			return (name, True)
		raise argparse.ArgumentTypeError("{!r} is not a legal bool value".format(value))
	elif type == "oracle":
		return (name, oracle.connect(value))
	elif type == "sqlite":
		return (name, sqlite.connect(value))
	elif type == "mysql":
		return (name, mysql.connect(value))
	elif type == "redis":
		return (name, redis.connect(value))
	elif type and type != "str":
		raise argparse.ArgumentTypeError("{!r} is not a legal type".format(type))
	return (name, value)


def print_exception_chain(exc):
	print("UL4 traceback (most recent call last):", file=sys.stderr)
	for exc in misc.exception_chain(exc):
		print()
		print(misc.format_exception(exc), file=sys.stderr)


def main(args=None):
	p = argparse.ArgumentParser(description="render UL4 templates with access to Oracle, MySQL, SQLite or Redis databases", epilog="For more info see http://www.livinglogic.de/Python/scripts/rul4.html")
	p.add_argument("templates", metavar="template", help="templates to be used", nargs="+")
	p.add_argument("-e", "--encoding", dest="encoding", help="Encoding for template sources (default %(default)s)", default="utf-8", metavar="ENCODING")
	p.add_argument("-w", "--whitespace", dest="whitespace", help="How to treat whitespace in template sources? (default %(default)s)", choices=("keep", "strip", "smart"), default="smart")
	p.add_argument("-t", "--stacktrace", dest="stacktrace", help="How to display stack traces in case of an error? (default %(default)s)", choices=("full", "short"), default="short")
	p.add_argument(      "--oracle", dest="oracle", help="Allow the templates to connect to Oracle databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--sqlite", dest="sqlite", help="Allow the templates to connect to SQLite databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--mysql", dest="mysql", help="Allow the templates to connect to MySQL databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--redis", dest="redis", help="Allow the templates to connect to Redis databases? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--system", dest="system", help="Allow the templates to execute system commands? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument(      "--import", dest="import_", help="Allow the templates to import templates from arbitrary paths? (default %(default)s)", action=misc.FlagAction, default=True)
	p.add_argument("-D", "--define", dest="defines", metavar="var=value", help="Pass additional parameters to the template (can be specified multiple times).", action="append", type=define)

	args = p.parse_args(args)

	templates = {}
	maintemplate = None
	for templatename in args.templates:
		if templatename == "-":
			templatestream = sys.stdin
			templatename = "stdin"
		else:
			templatestream = open(templatename, "r", encoding=args.encoding)
			templatename = os.path.basename(templatename)
			if os.path.extsep in templatename:
				templatename = templatename.rpartition(os.extsep)[0]
		templatesource = templatestream.read()
		templatename = fixname(templatename)
		if args.stacktrace == "short":
			try:
				template = ul4c.Template(templatesource, name=templatename, whitespace=args.whitespace)
			except Exception as exc:
				print_exception_chain(exc)
				return 1
		else:
			template = ul4c.Template(templatesource, name=templatename, whitespace=args.whitespace)
		# The first template is the main template
		if maintemplate is None:
			maintemplate = template
		templates[template.name] = template

	vars = dict(templates=templates, encoding=sys.stdout.encoding)
	if args.defines:
		vars.update(args.defines)
	if args.oracle:
		vars["oracle"] = oracle
	if args.mysql:
		vars["mysql"] = mysql
	if args.sqlite:
		vars["sqlite"] = sqlite
	if args.redis:
		vars["redis"] = redis
	if args.system:
		vars["system"] = system
	if args.import_:
		vars["import"] = import_
	if args.stacktrace == "short":
		try:
			for part in maintemplate.render(**vars):
				sys.stdout.write(part)
		except Exception as exc:
			print_exception_chain(exc)
			return 1
	else:
		for part in maintemplate.render(**vars):
			sys.stdout.write(part)


if __name__ == "__main__":
	sys.exit(main())



#! /usr/bin/env/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 1999-2015 by LivingLogic AG, Bayreuth/Germany
## Copyright 1999-2015 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


from ll.xist import xsc
from ll.xist.ns import html, php, chars, abbr


def test_basics_element():
	def check(pool, xmlns, xmlname, cls):
		assert pool.elementclass(xmlns, xmlname) is cls
		e = pool.element(xmlns, xmlname)
		assert (e.xmlns, e.xmlname) == (xsc.nsname(xmlns), xmlname)

	# empty pool
	pool = xsc.Pool()
	yield check, pool, html, "a", xsc.Element

	# register one element
	pool = xsc.Pool(html.a)
	yield check, pool, html, "a", html.a
	yield check, pool, html, "b", xsc.Element

	# register a module
	pool = xsc.Pool(html)
	yield check, pool, html, "a", html.a
	yield check, pool, html, "b", html.b
	yield check, pool, html, "c", xsc.Element


def test_basics_procinst():
	def check(pool, xmlname, content, cls):
		assert pool.procinstclass(xmlname) is cls
		procinst = pool.procinst(xmlname, content)
		assert procinst.xmlname == xmlname
		assert procinst.content == content

	pool = xsc.Pool(php.php)
	yield check, pool, "php", "foo", php.php
	yield check, pool, "nophp", "foo", xsc.ProcInst


def test_basics_entity():
	def check(pool, xmlname, cls):
		assert pool.entityclass(xmlname) is cls
		assert pool.entity(xmlname).xmlname == xmlname

	pool = xsc.Pool(abbr)
	yield check, pool, "xist", abbr.xist
	yield check, pool, "dontxist", xsc.Entity


def test_textcomment():
	pool = xsc.Pool()
	assert pool.text("foo") == xsc.Text("foo")
	assert pool.comment("foo") == xsc.Comment("foo")


def test_defaultpool():
	pool = xsc.threadlocalpool.pool
	assert pool.elementclass(html, "a") is html.a
	assert pool.procinstclass("php") is php.php
	assert pool.entityclass("xist") is abbr.xist
	assert pool.entityclass("euro") is chars.euro


def test_names():
	# Test classes where the Python and the XML name differ
	with xsc.Pool() as pool:
		class element1(xsc.Element):
			xmlname = "-element"
			xmlns = "nix"
		class procinst1(xsc.ProcInst):
			xmlname = "-procinst"
		class entity1(xsc.Entity):
			xmlname = "-entity"
		class charref1(xsc.CharRef):
			xmlname = "-charref"
			codepoint = 42
		class Attrs(xsc.Attrs):
			xmlns = "nix"
			class attr(xsc.TextAttr):
				xmlname = "-attr"
				xmlns = "nix"

	# elements
	assert pool.element1 is element1
	assert pool.elementclass("nix", "-element") is element1
	assert pool.element("nix", "-element") == element1()
	assert pool.elementclass("nix", "element1") is xsc.Element
	# make sure that the default pool didn't pick up the new class
	assert xsc.threadlocalpool.pool.elementclass("nix", "-element") is xsc.Element

	# procinsts
	assert pool.procinst1 is procinst1
	assert pool.procinstclass("-procinst") is procinst1
	assert pool.procinst("-procinst", "spam") == procinst1("spam")
	assert pool.procinstclass("procinst1") is xsc.ProcInst
	# make sure that the default pool didn't pick up the new class
	assert xsc.threadlocalpool.pool.procinstclass("-procinst") is xsc.ProcInst

	# entities
	assert pool.entity1 is entity1
	assert pool.entityclass("-entity") is entity1
	assert pool.entity("-entity") == entity1()
	assert pool.entityclass("entity1") is xsc.Entity
	# make sure that the default pool didn't pick up the new class
	assert xsc.threadlocalpool.pool.entityclass("-entity") is xsc.Entity
	# the charref is an entity too
	assert pool.entityclass("-charref") is charref1
	assert pool.entity("-charref") == charref1()
	assert pool.entityclass("charref1") is xsc.Entity
	# make sure that the default pool didn't pick up the new class
	assert xsc.threadlocalpool.pool.entityclass("-charref") is xsc.Entity

	# attributes
	assert pool.attrkey("nix", "-attr") is Attrs.attr
	assert pool.attrkey("nix", "attr") == ("nix", "attr")
	# make sure that the default pool didn't pick up the new class
	assert xsc.threadlocalpool.pool.attrkey("nix", "-attr") == ("nix", "-attr")


def test_names2():
	with xsc.Pool() as r:
		class el_(xsc.Element):
			xmlname = "el"
		class en_(xsc.Entity):
			xmlname = "en"
		class pi_(xsc.ProcInst):
			xmlname = "pi"
		class cr_(xsc.CharRef):
			xmlname = "cr"
			codepoint = 0x4242

	# Test elements
	assert set(r.elements()) == {el_}

	# Test entities
	assert set(r.entities()) == {cr_, en_}

	# Test procinsts
	assert set(r.procinsts()) == {pi_}


def test_stack():
	with xsc.Pool() as r1:
		class foo1(xsc.Element):
			xmlname = "foo"
			xmlns = "nix"
		with xsc.Pool() as r2:
			class foo2(xsc.Element):
				xmlname = "foo"
				xmlns = "nix"

	assert r1.elementclass("nix", "foo") is foo1
	assert r2.elementclass("nix", "foo") is foo2


def test_chain():
	with xsc.Pool() as p1:
		class foo1(xsc.Element):
			xmlname = "foo"
			xmlns = "nix"
		class baz(xsc.Element):
			xmlns = "nix"

	with xsc.Pool(p1) as p2:
		class foo2(xsc.Element):
			xmlname = "foo"
			xmlns = "nix"
		class bar(xsc.Element):
			xmlns = "nix"

	assert p1.elementclass("nix", "foo") is foo1
	assert p1.elementclass("nix", "bar") is xsc.Element
	assert p1.elementclass("nix", "baz") is baz

	assert p2.elementclass("nix", "foo") is foo2
	assert p2.elementclass("nix", "bar") is bar
	assert p2.elementclass("nix", "baz") is baz


def test_chain2():
	with xsc.Pool() as p1:
		class foo1(xsc.Element):
			xmlns = "nix"

	with xsc.Pool() as p2:
		class foo2(xsc.Element):
			xmlns = "nix"

	p = xsc.Pool(p1, p2)
	assert p.elementclass("nix", "foo2") is foo2


def test_mixedattrnames():
	with xsc.Pool() as r:
		class Attrs(xsc.Attrs):
			class a(xsc.TextAttr):
				xmlns = "test"
				xmlname = "A"
			class A(xsc.TextAttr):
				xmlns = "test"
				xmlname = "a"

		class Test(xsc.Element):
			xmlns = "test"
			class Attrs(xsc.Element.Attrs):
				class a(xsc.TextAttr):
					xmlname = "A"
				class A(xsc.TextAttr):
					xmlname = "a"

	xmlns = "test"

	node = Test(
		{
			Attrs.a: "a2",
			Attrs.A: "A2",
		},
		a="a",
		A="A"
	)

	def check(name, value):
		assert str(node[name]) == value
		assert str(node.attrs[name]) == value
		if isinstance(name, str):
			assert str(getattr(node.attrs, name)).swapcase() == value

	tests = [
		("A", "a"),
		("a", "A"),
		(Test.Attrs.a, "a"),
		(Test.Attrs.A, "A"),
		(Attrs.a, "a2"),
		(Attrs.A, "A2")
	]
	for (name, value) in tests:
		yield check, name, value


def test_xmlns():
	p = xsc.Pool(html)
	assert p.xmlns == html.xmlns


def test_itermethods():
	class pi(xsc.ProcInst):
		pass
	class en(xsc.Entity):
		pass
	class cr(xsc.CharRef):
		codepoint = 0x42

	p1 = xsc.Pool(html.a, pi, en, cr)
	p2 = xsc.Pool(html.b, p1)

	assert html.a in list(p2.elements())
	assert html.b in list(p2.elements())
	assert pi in list(p2.procinsts())
	assert en in list(p2.entities())
	assert cr in list(p2.entities())



#! /usr/bin/env/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 2000-2015 by LivingLogic AG, Bayreuth/Germany
## Copyright 2000-2015 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


from ll import ansistyle


def test_noswitch():
	def check(c):
		assert not ansistyle.switchcolor(c, c)

	for blink in (0000, 0o400):
		for underline in (0000, 0o200):
			for bold in (0000, 0o100):
				for foreground in (0000, 0o070):
					for background in (0000, 0o007):
						yield check, blink|underline|bold|foreground|background


def test_blink():
	assert str(ansistyle.Text(0o470, "foo")) == "\033[5mfoo\033[0m"
	assert str(ansistyle.Text(0o470, "foo", ansistyle.Text(0o070, "bar"))) == "\033[5mfoo\033[0mbar"


def test_underline():
	assert str(ansistyle.Text(0o270, "foo")) == "\033[4mfoo\033[0m"
	assert str(ansistyle.Text(0o270, "foo", ansistyle.Text(0o070, "bar"))) == "\033[4mfoo\033[0mbar"


def test_bold():
	assert str(ansistyle.Text(0o170, "foo")) == "\033[1mfoo\033[0m"
	assert str(ansistyle.Text(0o170, "foo", ansistyle.Text(0o070, "bar"))) == "\033[1mfoo\033[0mbar"


def test_background():
	assert str(ansistyle.Text(0o075, "foo")) == "\033[45mfoo\033[0m"
	assert str(ansistyle.Text(0o075, "foo", ansistyle.Text(0o070, "bar"))) == "\033[45mfoo\033[0mbar"


def test_foreground():
	assert str(ansistyle.Text(0o050, "foo")) == "\033[35mfoo\033[0m"
	assert str(ansistyle.Text(0o050, "foo", ansistyle.Text(0o070, "bar"))) == "\033[35mfoo\033[0mbar"


def test_all():
	assert str(ansistyle.Text(0o754, "foo")) == "\033[1;4;5;35;44mfoo\033[0m"
	assert str(ansistyle.Text(0o754, "foo", ansistyle.Text(0o070, "bar"))) == "\033[1;4;5;35;44mfoo\033[0mbar"


def test_nested():
	assert str(ansistyle.Text(0o075, "foo", ansistyle.Text(0o050, "bar"))) == "\033[45mfoo\033[35;40mbar\033[0m"
	assert str(ansistyle.Text(0o075, "foo", (0o050, "bar"))) == "\033[45mfoo\033[35;40mbar\033[0m"
	assert str(ansistyle.Text(0o075, "foo", (0o050, (0o060, "bar"), "baz"))) == "\033[45mfoo\033[36;40mbar\033[35mbaz\033[0m"


def test_bug():
	assert str(ansistyle.Text(-1, "x", (0o150, "y"), "z")) == "x\033[1;35my\033[0mz"


def test_escaped():
	class Escaped(ansistyle.EscapedText):
		def escapechar(self, c):
			if c == "?":
				return (0o060, "?")
			else:
				return c

	assert str(Escaped("!?!")) == "!\033[36m?\033[0m!"


def test_pushpop():
	c = ansistyle.Colorizer()
	c.pushcolor(0o060)
	assert "".join(c.feed("gurk")) == "\033[36mgurk"
	c.pushcolor(0o050)
	assert "".join(c.feed("hurz")) == "\033[35mhurz"
	c.popcolor()
	assert "".join(c.feed("hinz")) == "\033[36mhinz"
	c.popcolor()
	assert "".join(c.feed("kunz")) == "\033[0mkunz"




#!/usr/local/bin/python
# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 2007-2015 by LivingLogic AG, Bayreuth/Germany
## Copyright 2007-2015 by Walter Dörwald
##
## All Rights Reserved
##
## See ll/xist/__init__.py for the license


"""
Purpose
-------

``ucp`` is a script that copies files or directories. It is an URL-enabled
version of the ``cp`` system command. Via :mod:`ll.url` and :mod:`ll.orasql`
``ucp`` supports ``ssh`` and ``oracle`` URLs.


Options
-------

``ucp`` supports the following options:

	``urls``
		Two or more URLs. If more than two URLs are given or the last URL refers
		to an existing directory, the last URL is the target directory. All other
		sources are copied into this target directory. Otherwise one file is
		copied to another file.

	``-v``, ``--verbose`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Give a report during the copy process about the files copied and their
		sizes.

	``-c``, ``--color`` : ``yes``, ``no`` or ``auto``
		Should the output be colored? If ``auto`` is specified (the default) then
		the output is colored if stdout is a terminal.

	``-u``, ``--user``
		A user id or name. If given ``ucp`` will change the owner of the
		target files.

	``-g``, ``--group``
		A group id or name. If given ``ucp`` will change the group of the
		target files.

	``-r``, ``--recursive`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Copies files recursively.

	``-x``, ``--ignoreerrors`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Ignores errors occuring during the copy process (otherwise the copy
		process is aborted).

	``-i``, ``--include`` : pattern
		Only copy files that match the pattern.

	``-e``, ``--exclude`` : pattern
		Don't copy files that match the pattern.

	``--enterdir`` : pattern
		Only enter directories that match the pattern.

	``--skipdir`` : pattern
		Skip directories that match the pattern.

	``--ignorecase`` : ``false``, ``no``, ``0``, ``true``, ``yes`` or ``1``
		Perform case-insensitive pattern matching.


Examples
--------

Copy one file to another::

	$ ucp foo.txt bar.txt

Copy a file into an existing directory::

	$ ucp foo.txt dir/

Copy multiple files into a new or existing directory (and give a progress
report)::

	$ ucp foo.txt bar.txt baz.txt dir/ -v
	ucp: foo.txt -> dir/foo.txt (1,114 bytes)
	ucp: bar.txt -> dir/bar.txt (2,916 bytes)
	ucp: baz.txt -> dir/baz.txt (35,812 bytes)

Recursively copy the schema objects in an Oracle database to a local directory::

	ucp oracle://user:pwd@oracle.example.org/ db/ -r

Recursively copy the schema objects in an Oracle database to a remote directory::

	ucp oracle://user:pwd@oracle.example.org/ ssh://user@www.example.org/~/db/ -r
"""


import sys, re, argparse, contextlib

from ll import misc, url

try:
	import astyle
except ImportError:
	from ll import astyle

try:
	from ll import orasql # activate the oracle scheme
except ImportError:
	pass


__docformat__ = "reStructuredText"


def main(args=None):
	def copyone(urlread, urlwrite):
		strurlread = str(urlread)
		if urlread.isdir():
			if args.recursive:
				for u in urlread.walkfiles(include=args.include, exclude=args.exclude, enterdir=args.enterdir, skipdir=args.skipdir, ignorecase=args.ignorecase):
					copyone(urlread/u, urlwrite/u)
			else:
				if args.verbose:
					msg = astyle.style_default("ucp: ", astyle.style_url(strurlread), astyle.style_warn(" (directory skipped)"))
					stderr.writeln(msg)
		else:
			if args.verbose:
				msg = astyle.style_default("ucp: ", astyle.style_url(strurlread), " -> ")
				stderr.write(msg)
			try:
				with contextlib.closing(urlread.open("rb")) as fileread:
					with contextlib.closing(urlwrite.open("wb")) as filewrite:
						size = 0
						while True:
							data = fileread.read(262144)
							if data:
								filewrite.write(data)
								size += len(data)
							else:
								break
				if user or group:
					urlwrite.chown(user, group)
			except Exception as exc:
				if args.ignoreerrors:
					if args.verbose:
						exctype = misc.format_class(exc)
						excmsg = str(exc).replace("\n", " ").strip()
						msg = astyle.style_error(" (failed with {}: {})".format(exctype, excmsg))
						stderr.writeln(msg)
				else:
					raise
			else:
				if args.verbose:
					msg = astyle.style_default(astyle.style_url(str(urlwrite)), " ({:,} bytes)".format(size))
					stderr.writeln(msg)

	p = argparse.ArgumentParser(description="Copies URLs", epilog="For more info see http://www.livinglogic.de/Python/scripts/ucp.html")
	p.add_argument("urls", metavar="url", help="either one source and one target file, or multiple source files and one target dir", nargs="*", type=url.URL)
	p.add_argument("-v", "--verbose", dest="verbose", help="Be verbose? (default: %(default)s)", action=misc.FlagAction, default=False)
	p.add_argument("-c", "--color", dest="color", help="Color output (default: %(default)s)", default="auto", choices=("yes", "no", "auto"))
	p.add_argument("-u", "--user", dest="user", help="user id or name for target files")
	p.add_argument("-g", "--group", dest="group", help="group id or name for target files")
	p.add_argument("-r", "--recursive", dest="recursive", help="Copy stuff recursively? (default: %(default)s)", action=misc.FlagAction, default=False)
	p.add_argument("-x", "--ignoreerrors", dest="ignoreerrors", help="Ignore errors? (default: %(default)s)", action=misc.FlagAction, default=False)
	p.add_argument("-i", "--include", dest="include", metavar="PATTERN", help="Include only URLs matching PATTERN", action="append")
	p.add_argument("-e", "--exclude", dest="exclude", metavar="PATTERN", help="Exclude URLs matching PATTERN", action="append")
	p.add_argument(      "--enterdir", dest="enterdir", metavar="PATTERN", help="Only enter directories matching PATTERN", action="append")
	p.add_argument(      "--skipdir", dest="skipdir", metavar="PATTERN", help="Skip directories matching PATTERN", action="append")
	p.add_argument(      "--ignorecase", dest="ignorecase", help="Perform case-insensitive name matching? (default: %(default)s)", action=misc.FlagAction, default=False)

	args = p.parse_args(args)
	if len(args.urls) < 2:
		p.error("need at least one source url and one target url")
		return 1

	if args.color == "yes":
		color = True
	elif args.color == "no":
		color = False
	else:
		color = None
	stdout = astyle.Stream(sys.stdout, color)
	stderr = astyle.Stream(sys.stderr, color)

	user = args.user
	try:
		user = int(user)
	except (TypeError, ValueError):
		pass

	group = args.group
	try:
		group = int(group)
	except (TypeError, ValueError):
		pass

	with url.Context():
		urls = args.urls
		if len(urls) > 2 or urls[-1].isdir(): # treat target as directory
			for u in urls[:-1]:
				copyone(u, urls[-1]/u.file)
		else:
			copyone(urls[0], urls[-1])


if __name__ == "__main__":
	sys.exit(main())



# -*- coding: utf-8 -*-
# cython: language_level=3, always_allow_keywords=True

## Copyright 2012-2015 by LivingLogic AG, Bayreuth/Germany
## Copyright 2012-2015 by Walter Dörwald
##
## All Rights Reserved
##
## See LICENSE for the license

"""
Overview
========
The module/script :mod:`pysql` can be used to import data into an Oracle
database. It reads ``pysql`` files which are a variant of normal Oracle SQL
files.

A PySQL file can contain two different types of commands.


SQL commands
------------

A PySQL file may contains normal SQL commands. For the :mod:`pysql` script
to be able to execute these commands they must be terminated with a comment
line that starts with ``-- @@@``. :mod:`pysql` will strip off a trailing
``;`` or ``/`` from the command and execute it. Any exception that is raised
as a result of executing the command will stop the script and be reported.
This is in contrast to how ``sqlplus`` executes SQL command. ``sqlplus``
would continue after an error and exit with status code 0 even if there were
errors. It is also possible to explicitely ignore any exception raised by the
command by terminating the command with ``-- !!!`` (instead of ``-- @@@``).

A PySQL file that only contains SQL commands is still a valid SQL file from
the perspective of Oracle, so it still can be executed via ``sqlplus``.


PySQL commands
--------------

A pysql file may also contain PySQL commands. A PySQL command looks like a
Python dictionary literal. This literal must either be contained in a single
line or it must start with a line that only contains ``{`` and end at a
line that only contains ``}``.

For further information about the different commands and which keys they support,
see the class :class:`Command` and its subclasses.


Example
=======

The following is a complete PySQL file that will create a sequence, table and
procedure and will call the procedure to insert data into the table::

	create sequence person_seq
		increment by 10
		start with 10
		maxvalue 1.0e28
		minvalue 10
		nocycle
		cache 20
		noorder
	;

	-- @@@

	create sequence contact_seq
		increment by 10
		start with 10
		maxvalue 1.0e28
		minvalue 10
		nocycle
		cache 20
		noorder
	;

	-- @@@ person: table

	create table person
	(
		per_id integer not null,
		per_firstnane varchar2(200),
		per_lastnane varchar2(200)
	);

	-- @@@ person: primary key

	alter table person add constraint person_pk primary key(per_id);

	-- @@@ contact: table

	create table contact
	(
		con_id integer not null,
		per_id integer not null,
		con_type varchar2(200),
		con_value varchar2(200)
	);

	-- @@@ contact: primary key

	alter table contact add constraint contact_pk primary key(con_id);

	-- @@@ person: insert procedure

	create or replace procedure person_insert
	(
		c_user in varchar2,
		p_per_id in out integer,
		p_per_firstnane in varchar2 := null,
		p_per_lastnane in varchar2 := null
	)
	as
	begin
		if p_per_id is null then
			select person_seq.nextval into p_per_id from dual;
		end if;

		insert into person
		(
			per_id,
			per_firstnane,
			per_lastnane
		)
		values
		(
			p_per_id,
			p_per_firstnane,
			p_per_lastnane
		);
	end;
	/

	-- @@@ contact: insert procedure

	create or replace procedure contact_insert
	(
		c_user in varchar2,
		p_con_id in out integer,
		p_per_id in integer := null,
		p_con_type in varchar2 := null,
		p_con_value in varchar2 := null
	)
	as
	begin
		if p_con_id is null then
			select contact_seq.nextval into p_con_id from dual;
		end if;

		insert into contact
		(
			con_id,
			per_id,
			con_type,
			con_value
		)
		values
		(
			p_con_id,
			p_per_id,
			p_con_type,
			p_con_value
		);
	end;
	/

	-- @@@ import data

	{
		'type': 'procedure',
		'name': 'person_insert',
		'args': {
			'per_id': var('per_id_max'),
			'per_firstname': 'Max',
			'per_lastname': 'Mustermann',
		}
	}

	{
		'type': 'procedure',
		'name': 'contact_insert',
		'args': {
			'per_id': var('per_id_max'),
			'con_id': var('con_id_max'),
			'con_type': 'email',
			'con_value': 'max@example.org',
		}
	}

	{
		'type': 'file',
		'name': 'portrait_{per_id_max}.png',
		'content': b'\\x89PNG\\r\\n\\x1a\\n...',
	}

	{
		'type': 'resetsequence',
		'sequence': 'person_seq',
		'table': 'person',
		'field': 'per_id',
	}

	{"type": "compileall"}

	{"type": "checkerrors"}

This file can then be imported into an Oracle database with the following
command::

	python pysql.py user/pwd@database data.pysql

This will create two sequences, two tables and two procedures. Then it will
import two records, one by calling ``person_insert`` and one by calling
``contact_insert``. The PL/SQL equivalent of the above is::

	declare
		v_per_id_max integer;
		v_con_id_max integer;
	begin
		person_insert(
			per_id=v_per_id_max,
			per_firstname='Max',
			per_lastname='Mustermann'
		);
		contact_insert(
			con_id=v_con_id_max,
			per_id=v_per_id_max,
			con_type='email',
			con_value='max@example.org'
		)
	end;

Furthermore it will create one file (named something like ``portrait_42.png``)
and reset the sequence ``person_seq`` to the maximum value of the field
``per_id`` in the table ``person``.


Importing a PySQL file
======================

``pysql.py`` has no external dependencies except for :mod:`cx_Oracle` and can
be used as a script for importing a PySQL file into the database. As a script
it supports the following command line options:

	``connectstring``
		An Oracle connectstring.

	``file``
		The name of the PySQL file that will be read and imported. If ``file``
		isn't specified the commands are read from ``stdin``.

	``-v``, ``--verbose``
		Gives different levels of output while data is being imported to the
		database. Possible levels are: ``0`` (no output), ``1`` (one letter for
		each command), ``2`` (like ``1``, plus a summary of which command has been
		executed how often and which procedure has been called how often), ``3``
		(detailed output for each command/procedure call, plus summary)

	``-c``, ``--commit``
		Specifies when to commit database transactions. ``record`` commits after
		every command. ``once`` (the default) at the end of the script and
		``never`` rolls back the transaction after all imports.

	``-s``, ``--scpdirectory``
		The base directory for ``scp`` file copy commands. As files are copied
		via ``scp`` this can be a remote filename (like
		``root@www.example.org:~/uploads/``) and must include a trailing ``/``.

	``-f``, ``--filedirectory``
		The base directory for the ``file`` file save commands. It must include
		a trailing ``/``.

	``-t``, ``--terminator``
		The terminator after an SQL command (should be a valid SQL comment;
		default ``-- @@@``).

	``-T``, ``--terminatorignored``
		The error ignoring terminator after an SQL command (should be a valid
		SQL comment; default ``-- !!!``).
"""

# We're importing ``datetime``, so that it's available to ``eval()``
import sys, os, os.path, pwd, grp, argparse, operator, collections, datetime, tempfile, subprocess

import cx_Oracle


__docformat__ = "reStructuredText"


class Context:
	"""
	A :class:`Context` objects contains the configuration and run time information
	required for importing a PySQL file.
	"""
	def __init__(self, db=None, scpdirectory="", filedirectory="", commit="once", terminator="-- @@@", terminatorignore="-- !!!", verbose=0):
		self.keys = {}
		self.db = db
		self.cursor = db.cursor() if db is not None else None
		self.scpdirectory = scpdirectory
		self.filedirectory = filedirectory
		self.commit = commit
		self.terminator = terminator
		self.terminatorignore = terminatorignore
		self.verbose = verbose
		self.count = 0
		self.commandcounts = collections.Counter()
		self.procedurecounts = collections.Counter()
		self.errorcount = 0

	def load(self, stream):
		"""
		Load a PySQL file from :obj:`stream`. :obj:`stream` must be an iterable
		over lines that contain the PySQL commands.

		This function is a generator. Its output are the PySQL command objects
		(i.e. instances of :class:`Command`).
		"""
		reporterrors = True
		lines = []

		def makeblock():
			# Drop empty lines at the start
			while lines and not lines[0][1].strip():
				del lines[0]
			# Drop empty lines at the end
			while lines and not lines[-1][1].strip():
				del lines[-1]
			try:
				if lines:
					location = Location(stream.name, lines[0][0], lines[-1][0])
					text = "\n".join(line[1] for line in lines).strip()
					if text:
						args = {}
						lines.clear()
						if text.startswith("{") and text.endswith("}"):
							try:
								block = eval(text)
							except Exception as exc:
								raise Error(args) from exc
							args.update(block)
							if "type" not in args:
								args["type"] = "procedure"
						elif text.endswith((";", "/")):
							args["type"] = "sql"
							args["sql"] = text[:-1]
						else:
							raise ValueError("block terminator {!r} unknown".format(text[-1:]))

						type = args.pop("type")
						command = Command.fromdict(type, location, reporterrors, args)
						if isinstance(command, IncludeCommand):
							if not command.name:
								raise MissingKeyError("include", "name")
							filename = os.path.join(os.path.dirname(command.location.filename), command.name)
							with open(filename, "r", encoding="utf-8") as f:
								yield from self.load(f)
						else:
							yield command
			except Exception as exc:
				raise Error(location) from exc

		state = None # values ``"py"`` (inside PySQL block), ``"sql"`` (inside SQL block), ``None`` outside of any block
		for (i, line) in enumerate(stream, 1):
			line = line.rstrip()
			if state is None:
				if line.startswith("{"):
					lines.append((i, line))
					if line.endswith("}"):
						yield from makeblock()
					else:
						state = "py"
				elif line.startswith(self.terminator) or line.startswith(self.terminatorignore):
					pass # Still outside the block
				elif line.strip():
					lines.append((i, line))
					state = "sql"
			elif state == "py":
				lines.append((i, line))
				if line == "}": # A single unindented ``}``
					yield from makeblock()
					state = None
			elif state == "sql":
				if line.startswith(self.terminator):
					reporterrors = True
					yield from makeblock()
					state = None
				elif line.startswith(self.terminatorignore):
					reporterrors = False
					yield from makeblock()
					state = None
				else:
					lines.append((i, line))
		yield from makeblock()

	def executeall(self, commands):
		"""
		Execute all command in :obj:`commands`, which must be an iterable of
		:class:`Command` objects.
		"""
		try:
			if self.verbose == 2:
				print("commands:", end="", flush=True)
			for command in commands:
				try:
					command.execute(self)
				except Exception as exc:
					if command.reporterrors:
						if self.verbose == 2:
							print("(error)", flush=True)
						raise Error(command) from exc
					else:
						self.errorcount += 1
						if self.verbose >= 1:
							if self.verbose >= 3:
								exctext = str(exc).replace("\r\n", " ").replace("\r", " ").replace("\n", " ")
								print(" -> ignored {}.{}: {}".format(exc.__class__.__module__, exc.__class__.__qualname__, exctext))
							elif self.verbose >= 2:
								print("(error)", end="", flush=True)
				else:
					if isinstance(command, ProcedureCommand):
						self.procedurecounts[command.name] += 1
					self.commandcounts[command.type] += 1
					self.count += 1
			if self.commit == "once":
				self.db.commit()
			elif self.commit == "never":
				self.db.rollback()
		finally:
			if self.verbose >= 3:
				print()
		self._printsummary()

	def _printsummary(self):
		if self.verbose >= 2:
			commandcountvalues = self.commandcounts.values()
			l1 = len(str(max(commandcountvalues))) if commandcountvalues else 0
			l1 = max(l1, len(str(self.errorcount)))
			l2 = max(len(procname) for procname in self.procedurecounts) if self.procedurecounts else 0
			print()
			print("Summary")
			print("=======")
			anyoutput = False
			if self.commandcounts["procedure"]:
				anyoutput = True
				print("{:>{}} type".format("#", l1))
				print("{} {}".format("-"*l1, "-"*l2))
				for (procname, count) in sorted(self.procedurecounts.items(), key=operator.itemgetter(1)):
					print("{:>{}} {}".format(count, l1, procname))
				print("{} {}".format("-"*l1, "-"*l2))
			for cmdtype in ("procedure", "sql", "resetsequence", "setvar", "file", "scp", "compileall", "checkerrors"):
				if self.commandcounts[cmdtype]:
					anyoutput = True
					print("{:>{}} ({})".format(self.commandcounts[cmdtype], l1, cmdtype))
			if self.errorcount:
				print("{:>{}} error{} ignored".format(self.errorcount, l1, "s" if self.errorcount != 1 else ""))
			if not anyoutput:
				print("no commands executed")


###
### Command classes
###

class Command:
	"""
	The base class of all commands. A :class:`Command` object is created from
	a command dictionary literal in a PySQL file. The keys in the command
	dictorionary that are supported by all command types are the following:

		``type`` : string (optional)
			This is either ``"procedure"`` (the default), ``"sql"``, ``"file"``,
			``"scp"``, ``"resetsequence"``, ``"setvar"``, ``"include"``,
			``"compileall"`` or ``"checkerrors"`` and specifies the type of the
			PySQL command.

		``reporterrors`` : bool (optional)
			Specifies whether exceptions that happen during the execution of the
			command should be reported and terminate the script (``True``,
			the default), or should be ignored (``False``).
	"""
	def __init__(self, location, reporterrors):
		self.location = location
		self.reporterrors = reporterrors

	commands = {}

	@classmethod
	def fromdict(cls, type, location, reporterrors, d):
		if type in cls.commands:
			return cls.commands[type](location, reporterrors, **d)
		raise ValueError("command type {!r} unknown".format(type))

	def __str__(self):
		return "{} command {}".format(self.type, self.location)

	def executeload(self, value):
		if isinstance(value, load):
			value = value.execute(self.location.filename)
		return value

	def formatload(self, value):
		if isinstance(value, str):
			return "({:,} chars)".format(len(value))
		elif isinstance(value, bytes):
			return "({:,} bytes)".format(len(value))
		else:
			return repr(value)


def register(cls):
	Command.commands[cls.type] = cls
	return cls


@register
class IncludeCommand(Command):
	"""
	The ``"include"`` command includes another PySQL file. The filename is read
	from the key ``"name"``. This name is interpreted as being relative to the
	directory with the file containing the ``include`` command.
	"""
	type = "include"

	def __init__(self, location, reporterrors, name):
		super().__init__(location=location, reporterrors=reporterrors)
		self.name = name

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} name={0.name!r} {0.location} at {1:#x}>".format(self, id(self))

	def __str__(self):
		return "incluce command {}".format(self.location)


class _SQLCommand(Command):
	"""
	Common base class of :class:`ProcedureCommand` and :class:`SQLCommand`.
	"""
	def _executesql(self, context, query):
		queryargvars = {}
		for (argname, argvalue) in self.args.items():
			if isinstance(argvalue, var):
				if argvalue.key is None:
					queryargvars[argname] = None
				elif argvalue.key in context.keys:
					queryargvars[argname] = context.keys[argvalue.key]
				else:
					queryargvars[argname] = context.cursor.var(argvalue.type)
			elif isinstance(argvalue, sql):
				pass # no value
			elif isinstance(argvalue, load):
				queryargvars[argname] = self.executeload(argvalue)
			elif isinstance(argvalue, str) and len(argvalue) >= 4000:
				var_ = context.cursor.var(cx_Oracle.CLOB)
				var_.setvalue(0, argvalue)
				queryargvars[argname] = var_
			else:
				queryargvars[argname] = argvalue

		context.cursor.execute(query, queryargvars)

		newkeys = {}
		for (argname, argvalue) in self.args.items():
			if isinstance(argvalue, var) and argvalue.key is not None and argvalue.key not in context.keys:
				newkeys[argname] = context.keys[argvalue.key] = queryargvars[argname].getvalue(0)
		return newkeys

	def _formatargs(self, context):
		args = []
		if self.args:
			for (argname, argvalue) in self.args.items():
				if isinstance(argvalue, var) and argvalue.key in context.keys:
					arg = "{}={!r}={!r}".format(argname, argvalue, context.keys[argvalue.key])
				else:
					arg = "{}={!r}".format(argname, argvalue)
				args.append(arg)
		return ", ".join(args)


@register
class ProcedureCommand(_SQLCommand):
	"""
	A ``"procedure"`` command calls an Oracle procedure in the database.
	In addition to ``"type"`` and ``"reporterrors"`` the following keys are
	supported in the command dictionary:

		``name`` : string (required)
			The name of the procedure to be called (This may include ``.`` for
			calling a procedure in a package or one owned by a different user).

		``args`` : dictionary (optional)
			A dictionary with the names of the parameters as keys and the parameter
			values as values. PySQL supports all types as values that
			:mod:`cx_Oracle` supports. In addition to those, three special classes
			are supported:

			*	:class:`sql` objects can be used to specify that the paramater
				should be literal SQL. So e.g. ``sql("sysdate")`` will be the date
				when the PySQL script was executed.

			*	:class:`var` objects can be used to hold values that are ``OUT``
				parameters of the procedure. For example on first use of
				``var("foo_10")`` the value of the ``OUT`` parameter will be stored
				under the key ``"foo_10"``. The next time ``var("foo_10")`` is
				encountered the value stored under the key ``"foo_10"`` will be passed
				to the procedure. The type of the variable defaults to ``int``.
				If a different type is required it can be passed as the second
				argument to :class:`var`, e.g. ``var("foo_10", str)``.

			*	Finally :class:`load` objects can be used to load values from
				external files (as long as they are of type :class:`str` or
				:class:`bytes`). ``load("foo/bar.txt")`` will be replaced with the
				content of the external file ``foo/bar.txt`` (as a :class:`bytes`
				object). If a :class:`str` object is required, the encoding and
				error handler name can be passed to :class:`load` like this::

					load("foo/bar.txt", "utf-8", "replace")
	"""

	type = "procedure"

	def __init__(self, location, reporterrors, name, args=None):
		super().__init__(location=location, reporterrors=reporterrors)
		self.name = name
		self.args = args or {}

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} name={0.name!r} args={0.args!r} {0.location} at {1:#x}>".format(self, id(self))

	def __str__(self):
		return "procedure command {}".format(self.location)

	def _formatprocedurecall(self, context):
		return "{}({})".format(self.name, self._formatargs(context))

	def execute(self, context):
		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: procedure {}".format(context.count+1, self.location, self._formatprocedurecall(context)), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		queryargvalues = {}
		for (argname, argvalue) in self.args.items():
			if isinstance(argvalue, var):
				queryargvalues[argname] = ":{}".format(argname)
			elif isinstance(argvalue, sql):
				queryargvalues[argname] = argvalue.expression
				# no value
			elif isinstance(argvalue, str) and len(argvalue) >= 4000:
				queryargvalues[argname] = ":{}".format(argname)
			else:
				queryargvalues[argname] = ":{}".format(argname)

		query = "begin {}({}); end;".format(self.name, ", ".join("{}=>{}".format(*argitem) for argitem in queryargvalues.items()))
		result = self._executesql(context, query)

		if context.commit == "record":
			context.db.commit()

		if context.verbose >= 3:
			if result:
				print(" -> {}".format(", ".join("{}={!r}".format(argname, argvalue) for (argname, argvalue) in result.items())), flush=True)
			else:
				print(flush=True)
		return result


@register
class SQLCommand(_SQLCommand):
	"""
	An ``"sql"`` command directly executes an SQL statement in the Oracle database.
	In addition to ``"type"`` and ``"reporterrors"`` the following keys are
	supported in the command dictionary:

		``sql`` : string (required)
			The SQL to be executed. This may contain parameters in the form of
			``:paramname``. The values for those parameters will be taken from
			``args``.

		``args`` : dictionary (optional)
			A dictionary with the names of the parameters as keys and the parameter
			values as values. Similar to procedure calls :class:`var` and
			:class:`load` objects are supported. However :class:`sql` objects
			are not supported (they will be ignored).


	"""
	type = "sql"

	def __init__(self, location, reporterrors, sql, args=None):
		super().__init__(location=location, reporterrors=reporterrors)
		self.sql = sql
		self.args = args or {}

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} sql={0.sql!r} args={0.args!r} {0.location} at {1:#x}>".format(self, id(self))

	def __str__(self):
		return "sql command {}".format(self.location)

	def _formatsql(self, context):
		if self.args:
			return "{!r} with args {}".format(self.sql, self._formatargs(context))
		else:
			return repr(self.sql)

	def execute(self, context):
		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: sql {}".format(context.count+1, self.location, self._formatsql(context)), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		result = self._executesql(context, self.sql)

		if context.commit == "record":
			context.db.commit()

		if context.verbose >= 3:
			if result:
				print(" -> {}".format(", ".join("{}={!r}".format(argname, argvalue) for (argname, argvalue) in result.items())), flush=True)
			else:
				print(flush=True)

		return result


@register
class SetVarCommand(Command):
	"""
	The ``"setvar"`` command sets a variable to a fixed value. In addition to
	``"type"`` and ``"reporterrors"`` the following keys are supported in the
	command dictionary:

		``var``: string (required)
			The name of the variable to set.

		``value``: (required)
			The value of the variable.
	"""
	type = "setvar"

	def __init__(self, location, reporterrors, var, value):
		super().__init__(location=location, reporterrors=reporterrors)
		self.var = var
		self.value = value

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} var={0.var!r} value={0.value!r} {0.location} at {1:#x}>".format(self, id(self))

	def __str__(self):
		return "setvar command {}".format(self.location)

	def execute(self, context):
		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: set var {!r} to {!r}".format(context.count+1, self.location, self.var, self.value), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		context.keys[self.var] = self.value

		if context.verbose >= 3:
			print(" -> done", flush=True)


@register
class CheckErrorsCommand(Command):
	"""
	The ``"checkerrors"`` command checks that there are no compilation errors in
	the target schema. If there are, an exception will be raised. (The
	``reporterrors`` key is supported, but its value will be ignored).
	"""
	type = "checkerrors"

	def __init__(self, location, reporterrors):
		super().__init__(location=location, reporterrors=reporterrors)

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.location} at {1:#x}>".format(self, id(self))

	def __str__(self):
		return "checkerrors command {}".format(self.location)

	def execute(self, context):
		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: check errors".format(context.count+1, self.location), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		context.cursor.execute("select lower(type), name from user_errors group by lower(type), name")
		invalid_objects = [tuple(r) for r in context.cursor]

		if invalid_objects:
			raise CompilationError(invalid_objects)

		if context.verbose >= 3:
			print(" -> done", flush=True)


@register
class CompileAllCommand(Command):
	"""
	The ``"compileall"`` command will recompile all objects in the schema.
	"""
	type = "compileall"

	def __init__(self, location, reporterrors):
		super().__init__(location=location, reporterrors=reporterrors)

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} {0.location} at {1:#x}>".format(self, id(self))

	def __str__(self):
		return "compileall command {}".format(self.location)

	def execute(self, context):
		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: compile all".format(context.count+1, self.location), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		context.cursor.execute("begin dbms_utility.compile_schema(user); end;")

		if context.verbose >= 3:
			print(" -> done", flush=True)


@register
class SCPCommand(Command):
	"""
	The ``"scp"`` command creates a file by copying it via the ``scp`` command.
	In addition to ``"type"`` and ``"reporterrors"`` the following keys are
	supported in the command dictionary:

		``name`` : string (required)
			The name of the file to be created. It may contain ``format()`` style
			specifications containing any key that appeared in a ``"procedure"``
			or ``"sql"`` command. These specifiers will be replaced by the correct
			key values. As these files will be copied via the ``scp`` command,
			ssh file names can be used.

		``content``: bytes (required)
			The content of the file to be created. This can also be a
			:class:`load` object, to load the content from an external file.
	"""
	type = "scp"

	def __init__(self, location, reporterrors, name, content):
		super().__init__(location=location, reporterrors=reporterrors)
		self.name = name
		self.content = content

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} name={0.name!r} content={1} {0.location} at {2:#x}>".format(self, self.formatload(self.content), id(self))

	def __str__(self):
		return "scp command {}".format(self.location)

	def execute(self, context):
		filename = context.scpdirectory + self.name.format(**context.keys)

		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: scp {}".format(context.count+1, self.location, filename), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		content = self.executeload(self.content)
		with tempfile.NamedTemporaryFile(delete=False) as f:
			f.write(content)
			tempname = f.name
		try:
			result = subprocess.run(["scp", "-q", tempname, filename], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
			if result.returncode:
				raise SCPError(result.returncode, (result.stdout or result.stderr).decode(errors="replace"))
		finally:
			os.remove(tempname)

		if context.verbose >= 3:
			print(" -> {} bytes written".format(len(self.content)), flush=True)


@register
class FileCommand(Command):
	"""
	The ``"file"`` command creates a file by directly saving it from Python.
	In addition to ``"type"`` and ``"reporterrors"`` the following keys are
	supported in the command dictionary:

		``name`` : string (required)
			The name of the file to be created. It may contain ``format()`` style
			specifications containing any key that appeared in a ``"procedure"`` or
			``"sql"`` command. These specifiers will be replaced by the correct
			key values.

		``content``: bytes (required)
			The content of the file to be created. This can also be a
			:class:`load` object, to load the content from an external file.

		``mode``: integer (optional)
			The file mode for the new file. If the mode is specified :func:`os.chmod`
			will be called on the file.

		``owner``: integer or string (optional)
			The owner of the file (as a user name or a uid).

		``group``: integer or string (optional)
			The owning group of the file (as a group name or a gid).
			If ``owner`` or ``group`` is given, :func:`os.chown` will be called on
			the file.
	"""
	type = "file"

	def __init__(self, location, reporterrors, name, content, mode=None, owner=None, group=None):
		super().__init__(location=location, reporterrors=reporterrors)
		self.name = name
		self.content = content
		self.mode = mode
		self.owner = owner
		self.group = group

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} name={0.name!r} content={1} {0.location} at {2:#x}>".format(self, self.formatload(self.content), id(self))

	def __str__(self):
		return "file command {}".format(self.location)

	def execute(self, context):
		filename = context.filedirectory + self.name.format(**context.keys)

		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: file {}".format(context.count+1, self.location, filename), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		content = self.executeload(self.content)

		try:
			with open(filename, "wb") as f:
				f.write(content)
		except FileNotFoundError: # probably the directory doesn't exist
			(splitpath, splitname) = os.path.split(filename)
			if splitpath:
				os.makedirs(splitpath)
				with open(filename, "wb") as f:
					f.write(content)
			else:
				raise # we don't have a directory to make so pass the error on

		if self.mode:
			os.chmod(filename, self.mode)
		if self.owner or self.group:
			if self.owner:
				uid = self.owner
				if isinstance(uid, str):
					uid = pwd.getpwnam(uid)[2]
			else:
				uid = -1
			if self.group:
				gid = self.group
				if isinstance(gid, str):
					gid = grp.getgrnam(gid)[2]
			else:
				gid = -1
			os.chown(filename, uid, gid)

		if context.verbose >= 3:
			msg = " -> {:,} bytes written".format(len(content))
			optionmsg = []
			if self.mode:
				optionmsg.append("mode {:#o}".format(self.mode))
			if self.owner:
				optionmsg.append("owner {!r}".format(self.owner))
			if self.group:
				optionmsg.append("group {!r}".format(self.group))
			if optionmsg:
				msg = "{} ({})".format(msg, ", ".join(optionmsg))
			print(msg, flush=True)


@register
class ResetSequenceCommand(Command):
	"""
	The ``"resetsequence"`` command resets a sequence in the Oracle database to the
	maximum value of a field in a table. In addition to ``"type"`` and
	``"reporterrors`` the following keys are supported in the command dictionary:

		``sequence``: string (required)
			The name of the sequence to reset.

		``table``: string (required)
			The name of the table that contains the field.

		``field``: string (required)
			The name of the field in the table ``table``. The sequence will be
			reset to a value, so that fetching the next value from the sequence
			will deliver a value that is larger than the maximum value of the field
			``field`` in the table ``table``.

		``minvalue``: integer (optional, default taken from sequence)
			The minimum value for the sequence.

		``increment``: integer (optional, default taken from sequence)
			The increment (i.e. the stop size) for the sequence.
	"""
	type = "resetsequence"

	def __init__(self, location, reporterrors, sequence, table, field, minvalue=None, increment=None):
		super().__init__(location=location, reporterrors=reporterrors)
		self.sequence = sequence
		self.table = table
		self.field = field
		self.minvalue = minvalue
		self.increment = increment

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} sequence={0.sequence!r} {0.location} at {1:#x}>".format(self, id(self))

	def __str__(self):
		return "resetsequence command {}".format(self.location)

	def execute(self, context):
		if context.verbose >= 1:
			if context.verbose >= 3:
				print("#{:,} {}: resetting sequence {} to maximum value from {}.{}".format(context.count+1, self.location, self.sequence, self.table, self.field), end="", flush=True)
			else:
				print(" " + self.type, end="", flush=True)

		# Fetch information about the sequence
		context.cursor.execute("select min_value, increment_by, last_number from user_sequences where lower(sequence_name)=lower(:name)", name=self.sequence)
		oldvalues = context.cursor.fetchone()
		if oldvalues is None:
			raise ValueError("sequence {!r} unknown".format(self.sequence))
		increment = self.increment
		if increment is None:
			increment = oldvalues[1]
		minvalue = self.minvalue
		if minvalue is None:
			minvalue = oldvalues[0]
		context.cursor.execute("select {}.nextval from dual".format(self.sequence))
		seqvalue = context.cursor.fetchone()[0]

		# Fetch information about the table values
		context.cursor.execute("select nvl(max({}), 0) from {}".format(self.field, self.table))
		tabvalue = context.cursor.fetchone()[0]

		step = max(tabvalue, minvalue) - seqvalue
		if step:
			context.cursor.execute("alter sequence {} increment by {}".format(self.sequence, step))
			context.cursor.execute("select {}.nextval from dual".format(self.sequence))
			seqvalue = context.cursor.fetchone()[0]
			context.cursor.execute("alter sequence {} increment by {}".format(self.sequence, increment))
			if context.verbose >= 3:
				print(" -> reset to {}".format(seqvalue), flush=True)
			return seqvalue
		else:
			if context.verbose >= 3:
				print(" -> no reset required", flush=True)
			return None


###
### Classes to be used by the PySQL commands
###

class var:
	"""
	:class:`var` instances are used to mark procedure values that are ``OUT``
	parameters. On first use the parameter is used as an ``OUT`` parameter and
	PySQL will remembers the OUT value under the unique key specified in the
	constructor. When a :class:`var` object is used a second time its value will
	be passed to the procedure as a normal ``IN`` parameter.
	"""

	def __init__(self, key, type=int):
		"""
		Create a :class:`var` instance. :obj:`key` is a unique name for the value.
		:obj:`type` is the type of the value (defaulting to :class:`int`).
		"""
		self.key = key
		self.type = type

	def __repr__(self):
		if self.type is int:
			fmt = "var({0.key!r})"
		elif self.type.__module__ == "builtins":
			fmt = "var({0.key!r}, {0.type.__qualname__})"
		else:
			fmt = "var({0.key!r}, {0.type.__module__}.{0.type.__qualname__})"
		return fmt.format(self)


class sql:
	"""
	An :class:`sql` object can be used to specify an SQL expression as a
	procedure parameter instead of a fixed value. For example passing the current
	date (i.e. the date of the import) can be done with ``sql("sysdate")``.
	"""

	def __init__(self, expression):
		self.expression = expression

	def __repr__(self):
		return "sql({!r})".format(self.expression)


class load:
	"""
	A :class:`load` object can be used to load a :class:`str` or :class:`bytes`
	object from an external file.
	"""

	def __init__(self, filename, encoding=None, errors=None):
		"""
		Create a new :class:`load` object. :obj:`filename` is the name of the file
		to be loaded. The filename is treated as being relative to the directory
		containing the pysql file that contains the PySQL command with the
		:class:`load` object. If :obj:`encoding` is ``None`` the content of the
		file will be returned as a :class:`bytes` object, otherwise :obj:`encoding`
		and :obj:`errors` will be used for decoding the file content into a string.
		"""
		self.filename = filename
		self.encoding = encoding
		self.errors = errors

	def __repr__(self):
		fmt = "load({0.filename!r}"
		if self.encoding is not None:
			fmt += ", encoding={0.encoding!r}"
		if self.errors is not None:
			fmt += ", errors={0.errors!r}"
		fmt += ")"
		return fmt.format(self)

	def execute(self, basefilename):
		"""
		Read the file and return the file content as a :class:`bytes` or
		:class:`str` object. :obj:`basefilename` is the filename containing the
		PySQL command with the :class:`load` object (i.e. this determines the
		base directory).
		"""
		filename = os.path.join(os.path.dirname(basefilename), self.filename)
		if self.encoding is None: # load bytes
			with open(filename, "rb") as f:
				return f.read()
		else: # load string
			with open(filename, "r", encoding=self.encoding, errors=self.errors) as f:
				return f.read()


###
### Exception classes and location information
###

class Error(Exception):
	def __init__(self, location):
		self.location = location

	def __str__(self):
		return str(self.location)


class CompilationError(Exception):
	"""
	Exception raised by :class:`CheckErrorsCommand` when invalid database
	objects are encountered.
	"""
	def __init__(self, objects):
		self.objects = objects

	def __str__(self):
		if len(self.objects) == 1:
			return "one invalid db object: {} {}".format(*self.objects[0])
		else:
			return "{:,} invalid db objects: {}".format(len(self.objects), ", ".join("{} {}".format(*object) for object in self.objects))


class SCPError(Exception):
	"""
	Exception raised by :class:`SCPCommand` when a call to the ``scp`` comamnd
	fails.
	"""
	def __init__(self, status, msg):
		self.status = status
		self.msg = msg

	def __str__(self):
		return "scp failed with code {}: {}".format(self.status, self.msg)


class Location:
	"""
	The location of a PySQL/SQL command in a pysql file.
	"""
	def __init__(self, filename, startline, endline):
		self.filename = filename
		self.startline = startline
		self.endline = endline

	def __repr__(self):
		return "<{0.__class__.__module__}.{0.__class__.__qualname__} filename={0.filename!r} startline={0.startline!r} endline={0.endline!r} at {1:#x}>".format(self, id(self))

	def __str__(self):
		if self.startline is None and self.endline is None:
			return "in {!r}".format(self.filename)
		elif self.startline == self.endline:
			return "in {!r} at line {:,}".format(self.filename, self.startline)
		else:
			return "in {!r} at lines {:,}-{:,}".format(self.filename, self.startline, self.endline)


def main(args=None):
	p = argparse.ArgumentParser(description="Import a pysql file into an Oracle database", epilog="For more info see http://www.livinglogic.de/Python/pysql/index.html")
	p.add_argument("connectstring", help="Oracle connect string")
	p.add_argument("file", nargs="?", help="Name of the pysql file (default: read from stdin)", type=argparse.FileType("r"), default=sys.stdin)
	p.add_argument("-v", "--verbose", dest="verbose", help="Give a progress report? (default %(default)s)", type=int, default=2, choices=(0, 1, 2, 3))
	p.add_argument("-c", "--commit", dest="commit", help="When should database transactions be committed? (default %(default)s)", default="once", choices=("record", "once", "never"))
	p.add_argument("-s", "--scpdirectory", dest="scpdirectory", metavar="DIR", help="File name prefix for files to be copied via the 'scp' command (default: current directory)", default="")
	p.add_argument("-f", "--filedirectory", dest="filedirectory", metavar="DIR", help="File name prefix for files to be copied via the 'file' command (default: current directory)", default="")
	p.add_argument("-t", "--terminator", dest="terminator", metavar="STRING", help="Terminator after an SQL command (should be a valid SQL comment; default %(default)r)", default="-- @@@")
	p.add_argument("-T", "--terminatorignore", dest="terminatorignore", metavar="STRING", help="Terminator after an SQL command that ignores errors (should be a valid SQL comment; default %(default)r)", default="-- !!!")

	args = p.parse_args(args)

	context = Context(
		db=cx_Oracle.connect(args.connectstring),
		scpdirectory=args.scpdirectory,
		filedirectory=args.filedirectory,
		commit=args.commit,
		terminator=args.terminator,
		terminatorignore=args.terminatorignore,
		verbose=args.verbose,
	)
	context.executeall(context.load(args.file))


if __name__ == "__main__":
	sys.exit(main())
